/***************************************************************************\
  "re.h" - Runtime Engine (Phase E)
  Revision 17; 10 Nov 2002 - 4 Feb 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef RE_H
#define RE_H

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

// Local //
#include "global.h"
#include "error.h"
#include "icm.h"
#include "k_string.h"
#include "rstack.h"
#include "global_conv.h"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// Internal //
typedef enum
{
  fioInput, fioOutput
} FIOMode;

typedef struct FileElement_tag
{
  FILE *file;
  FIOMode mode;
} FileElement;

FileElement *mFiles[GLOBAL_RE_MAX_FILE_HANDLES];
Statement *mSM;
RStack *mLocalStack, *mCallStack, *mReturnStack;
TAC mTAC;


// External //
void re_init ();
void re_deinit ();
void re_run ();

// Internal - Run //
void mRunStatement ();
void mRunJump ();
void mRunCall ();
void mRunRet ();
void mRunConditional ();
void mRunConversion ();
void mRunAssignment ();
void mRunStackOperation ();
void mRunUnaryArith ();
void mRunBinaryArith ();
void mRunBinaryString ();
void mRunConsoleInput ();
void mRunConsoleOutput ();
void mRunASCIIConversion ();
void mRunRandom ();
void mRunFileIO ();

// Internal - Special //
void mGetStatement ();

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

//// re_init
//// Initialize re module
void re_init ()
{
  int i;

  for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
  {
    mFiles[i] = NULL;
  }

  mLocalStack = rstack_new ();
  mCallStack = rstack_new ();
  mReturnStack = rstack_new ();

  srand((unsigned)time(NULL));
}

//// re_deinit
//// Deinitialize re module
void re_deinit ()
{
  rstack_del (mLocalStack);
  rstack_del (mCallStack);
  rstack_del (mReturnStack);
}

//// re_run
void re_run ()
{
  mGetStatement();
  gnCurrentLine = 0;

  while (mTAC != tacSTOP)
  {
    mRunStatement ();
    mGetStatement ();
  }
}

/////////
// Run //
/////////

//// mRunStatement
void mRunStatement ()
{
  switch (mTAC)
  {
    case (tacNOP):
      // do nothing! //
      break;
    case (tacLINE):
      gnCurrentLine = mSM->number;
      JFREE(gsCurrentFilename);
      gsCurrentFilename = k_clone(mSM->a->value);
      break;
    case (tacJUMP):
      mRunJump ();
      break;
    case (tacCALL):
      mRunCall ();
      break;
    case (tacRET):
      mRunRet ();
      break;
    case (tacCOND):
      mRunConditional ();
      break;
    case (tacXIR):
    case (tacXIS):
    case (tacXRI):
    case (tacXRS):
    case (tacXSI):
    case (tacXSR):
      mRunConversion ();
      break;
    case (tacSET):
      mRunAssignment ();
      break;
    case (tacPUSHL):
    case (tacPOPL):
    case (tacPUSHR):
    case (tacPOPR):
      mRunStackOperation ();
      break;
    case (tacNEG):
      mRunUnaryArith ();
      break;
    case (tacADD):
    case (tacSUB):
    case (tacMULT):
    case (tacDIV):
    case (tacMOD):
    case (tacPOW):
    case (tacEQ):
    case (tacNEQ):
    case (tacLT):
    case (tacGT):
    case (tacLTEQ):
    case (tacGTEQ):
    case (tacOR):
    case (tacAND):
      mRunBinaryArith ();
      break;
    case (tacCAT):
    case (tacSEQ):
    case (tacSNEQ):
      mRunBinaryString ();
      break;
    case (tacCIN):
      mRunConsoleInput ();
      break;      
    case (tacCOUT):
      mRunConsoleOutput ();
      break;
    case (tacASC):
      mRunASCIIConversion ();
      break;
    case (tacFIO):
      mRunFileIO ();
      break;
    case (tacRAND):
      mRunRandom ();
      break;    
    default:
      err (true, ER_INVALID_INSTRUCTION, gvsTACStrings[mTAC]);
      erre (EFF_TERMINATE);
      break;
  }
}

//// mRunJump
void mRunJump ()
{
  // FORMAT: jump (index) (null) (null) 0 //

  if (mSM->a->constant == true)
  {
    icm_goto (*(int*)(mSM->a->value));
  }
  else
  {
    err (true, ER_JUMP_NOT_DEFINED, mSM->a->key);
    erre (EFF_TERMINATE);
  }
}

//// mRunCall
void mRunCall ()
{
  // FORMAT: call (tRoutine) (null) (null) 0 //

  Symbol *return_index;

  // push return address onto the runtime stack //
  return_index = sym_new ("return_index");
  sym_set (return_index, repLiteral, tLabel, false, &gxCurr);

  //scope_add_symbol_to_list (return_index); // add symbol to deletion list

  rstack_push (mCallStack, return_index);

  sym_del (return_index);

  // jump to the fisrt instruction of the routine //
  icm_goto ( ((Routine *)(mSM->a->value))->location );
}

//// mRunRet
void mRunRet ()
{
  // FORMAT: ret (null) (null) (null) 0 //

  void *return_value;

  // type is a label //
  if (rstack_peektype(mCallStack) == tLabel)
  {
    return_value = rstack_pop (mCallStack);
    
    if (return_value == NULL)
    {
      err (true, ER_RET_POP_STACK_EMPTY, NULL);
      erre (EFF_TERMINATE);
    }
    else
    {
      // jump to the next instruction of the calling routine //
      icm_goto (*(int*)(return_value));
    }

    JFREE(return_value);
  }
  // type is not label //
  else
  {
    err (true, ER_RET_POP_NOT_LABEL_TYPE, gvsTypeStrings[rstack_peektype(mCallStack)]);
    erre (EFF_TERMINATE);
  }
}

//// mRunConditional
void mRunConditional ()
{
  // FORMAT: cond (boolean) (true destination) (false destination) 0 //

  if ( (*(int*)(mSM->a->value)) != 0 )
  {
    icm_goto (*(int*)(mSM->b->value));
  }
  else
  {
    icm_goto (*(int*)(mSM->c->value));
  }
}

//// mRunConversion
void mRunConversion ()
{
  // FORMAT: x__ (dest) (source) (null) 0 //

  int ntemp;
  float ftemp;
  char *stemp;

  ntemp = 0;
  ftemp = 0.0;
  stemp = NULL;

  JFREE(mSM->a->value); // free the current value

  switch (mTAC)
  {
    case (tacXIR):
      ftemp = (float) *(int*)mSM->b->value;
      mSM->a->value = k_new_f (ftemp);
      break;
    case (tacXIS):
      stemp = conv_is ( *(int*)mSM->b->value );
      mSM->a->value = stemp;
      break;
    case (tacXRI):
      ntemp = (int) *(float*)mSM->b->value;
      mSM->a->value = k_new_n (ntemp);
      break;
    case (tacXRS):
      stemp = conv_fs ( *(float*)mSM->b->value );
      mSM->a->value = stemp;
      break;
    case (tacXSI):
      ntemp = conv_si ( (char*) mSM->b->value );
      mSM->a->value = k_new_n (ntemp);
      break;
    case (tacXSR):
      ftemp = conv_sf ( (char*) mSM->b->value );
      mSM->a->value = k_new_f (ftemp);
      break;
  }
}

//// mRunAssignment
void mRunAssignment ()
{
  // FORMAT: set (dest) (source) (null) 0 //

  int ntemp;
  float ftemp;

  ntemp = 0;
  ftemp = 0.0;

  JFREE(mSM->a->value); // free the current value

  // Put a copy of B's value in A //
  switch (mSM->a->type)
  {
    case (tInt):
      if (mSM->b->type == tRat) ntemp = (int) *(float*)mSM->b->value;
      else                      ntemp = *(int*)mSM->b->value;
      mSM->a->value = k_new_n (ntemp);
      break;
    case (tRat):
      if (mSM->b->type == tInt) ftemp = (float) *(int*)mSM->b->value;
      else                      ftemp = *(float*)mSM->b->value;
      mSM->a->value = k_new_f (ftemp);
      break;
    case (tStr):
      mSM->a->value = k_clone (mSM->b->value);
      break;
  }
}

//// mRunStackOperation
void mRunStackOperation ()
{
  // PUSHL //
  if (mTAC == tacPUSHL)
  {
    rstack_push (mLocalStack, mSM->a);
  }

  // POPL //
  else if (mTAC == tacPOPL)
  {
    // types are the same //
    if (mSM->a->type == rstack_peektype(mLocalStack))
    {
      JFREE(mSM->a->value);

      mSM->a->value = rstack_pop (mLocalStack);

      if ((mSM->a->value == NULL) && (mSM->a->type != tStr)) // only error if not str, strings can be NULL
      {
        err (true, ER_POP_STACK_EMPTY, NULL);
        erre (EFF_TERMINATE);
      }
    }
    // types are different //
    else
    {
      if (rstack_peektype(mLocalStack) == tNull)
      {
        err (true, ER_POP_STACK_EMPTY, NULL);
        erre (EFF_TERMINATE);
      }
      else
      {
        err (true, ER_POP_MISMATCHED_TYPE, gvsTypeStrings[rstack_peektype(mLocalStack)]);
        erre (EFF_TERMINATE);
      }
    }
  }

  // PUSHR //
  if (mTAC == tacPUSHR)
  {
    rstack_push (mReturnStack, mSM->a);
  }

  // POPR //
  else if (mTAC == tacPOPR)
  {
    // types are the same //
    if (mSM->a->type == rstack_peektype(mReturnStack))
    {
      JFREE(mSM->a->value);

      mSM->a->value = rstack_pop (mReturnStack);

      if ((mSM->a->value == NULL) && (mSM->a->type != tStr)) // only error if not str, strings can be NULL
      {
        err (true, ER_POP_STACK_EMPTY, NULL);
        erre (EFF_TERMINATE);
      }
    }
    // types are different //
    else
    {
      if (rstack_peektype(mReturnStack) == tNull)
      {
        err (true, ER_POP_STACK_EMPTY, NULL);
        erre (EFF_TERMINATE);
      }
      else
      {
        err (true, ER_POP_MISMATCHED_TYPE, gvsTypeStrings[rstack_peektype(mReturnStack)]);
        erre (EFF_TERMINATE);
      }
    }
  }
}

//// mRunUnaryArith
void mRunUnaryArith ()
{
  // FORMAT: neg (dest) (source) (null) 0 //

  int ntemp;
  float ftemp;

  ntemp = 0;
  ftemp = 0.0;

  JFREE(mSM->a->value); // free the current value

  // Perform the operation //
  switch (mSM->a->type)
  {
    case (tInt):
      ntemp = *(int*)mSM->b->value;
      mSM->a->value = k_new_n (-ntemp);
      break;
    case (tRat):
      ftemp = *(float*)mSM->b->value;
      mSM->a->value = k_new_f (-ftemp);
      break;
  }
}

//// mRunBinaryArith
void mRunBinaryArith ()
{
  // FORMAT: ___ (dest) (source_b) (source_c) 0 //

  double db, dc, dresult;
  float fb, fc, fresult;

  // calculations will be in integers
  if (mSM->a->type == tInt)
  {
    // Put the double value of B in 'db' //
    if (mSM->b->type != tRat)
    {
      db = (double) *(int*)mSM->b->value;
    }
    else
    {
      db = *(double*)mSM->b->value;
    }

    // Put the double value of C in 'dc' //
    if (mSM->c->type != tRat)
    {
      dc = (double) *(int*)mSM->c->value;
    }
    else
    {
      dc = *(double*)mSM->c->value;
    }

    switch (mTAC)
    {
      case (tacADD):
        dresult = db + dc;
        break;
      case (tacSUB):
        dresult = db - dc;
        break;
      case (tacMULT):
        dresult = db * dc;
        break;
      case (tacDIV):
        if (dc == 0.0) {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        } else  {
          dresult = db / dc;
        }
        break;
      case (tacMOD):
        if (dc == 0.0) {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        } else  {
          dresult = (double) ((int)(db) % (int)(dc));
        }
        break;
      case (tacPOW):
        dresult = (double) pow (db, dc);
        break;
      case (tacEQ):
        dresult = (double) (db == dc);
        break;
      case (tacNEQ):
        dresult = (double) (db != dc);
        break;
      case (tacLT):
        dresult = (double) (db < dc);
        break;
      case (tacGT):
        dresult = (double) (db > dc);
        break;
      case (tacLTEQ):
        dresult = (double) (db <= dc);
        break;
      case (tacGTEQ):
        dresult = (double) (db >= dc);
        break;
      case (tacOR):
        dresult = (double) (db || dc);
        break;
      case (tacAND):
        dresult = (double) (db && dc);
        break;
    }

    JFREE(mSM->a->value); // free the current value

    // Put the value of 'dresult' into A //
    mSM->a->value = k_new_n ( (int) dresult );
  }

  // calculations will be in rationals
  else
  {
    // Put the float value of B in 'fb' //
    if (mSM->b->type != tRat)
    {
      fb = (float) *(int*)mSM->b->value;
    }
    else
    {
      fb = *(float*)mSM->b->value;
    }

    // Put the float value of C in 'fc' //
    if (mSM->c->type != tRat)
    {
      fc = (float) *(int*)mSM->c->value;
    }
    else
    {
      fc = *(float*)mSM->c->value;
    }

    switch (mTAC)
    {
      case (tacADD):
        fresult = fb + fc;
        break;
      case (tacSUB):
        fresult = fb - fc;
        break;
      case (tacMULT):
        fresult = fb * fc;
        break;
      case (tacDIV):
        if (fc == 0.0) {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        } else  {
          fresult = fb / fc;
        }
        break;
      case (tacMOD):
        if (fc == 0.0) {
          err (true, ER_DIV_BY_ZERO, NULL);
          erre (EFF_TERMINATE);
        } else  {
          fresult = (float) ((int)(fb) % (int)(fc));
        }
        break;
      case (tacPOW):
        fresult = (float) pow (fb, fc);
        break;
      case (tacEQ):
        fresult = (float) (fb == fc);
        break;
      case (tacNEQ):
        fresult = (float) (fb != fc);
        break;
      case (tacLT):
        fresult = (float) (fb < fc);
        break;
      case (tacGT):
        fresult = (float) (fb > fc);
        break;
      case (tacLTEQ):
        fresult = (float) (fb <= fc);
        break;
      case (tacGTEQ):
        fresult = (float) (fb >= fc);
        break;
      case (tacOR):
        fresult = (float) (fb || fc);
        break;
      case (tacAND):
        fresult = (float) (fb && fc);
        break;
    }

    JFREE(mSM->a->value); // free the current value

    // Put the value of 'fresult' into A //
    mSM->a->value = k_new_f ( (float) fresult );
  }
}

//// mRunBinaryString
void mRunBinaryString ()
{
  char *temp;

  // FORMAT: cat (dest) (source_b) (source_c) 0 //
  if (mTAC == tacCAT)
  {
    temp = k_new (strlen (mSM->b->value) + strlen(mSM->c->value) + 1);
    k_copy (temp, "");
    k_cat (temp, mSM->b->value);
    k_cat (temp, mSM->c->value);

    k_del (mSM->a->value);
    mSM->a->value = temp;
  }
  else if (mTAC == tacSEQ)
  {
    JFREE(mSM->a->value);
    if (strcmp ((char*)(mSM->b->value), (char*)(mSM->c->value)) == 0)
      mSM->a->value = k_new_f (1.0);
    else
      mSM->a->value = k_new_f (0.0);
  }
  else if (mTAC == tacSNEQ)
  {
    JFREE(mSM->a->value);
    if (strcmp ((char*)(mSM->b->value), (char*)(mSM->c->value)) != 0)
      mSM->a->value = k_new_f (1.0);
    else
      mSM->a->value = k_new_f (0.0);
  }
}

//// mRunConsoleInput
void mRunConsoleInput ()
{
  char input[GLOBAL_RUNTIME_ENGINE_CIN_LINE_LEN];

  JFREE(mSM->a->value); // free the current value

  // Loop until the user inputs a non-blank line //
  gets (input);
  while (strlen(input) == 0) gets (input);

  switch (mSM->a->type)
  {
    case (tInt):
      mSM->a->value = k_new_n ( conv_si (input) );
      break;
    case (tRat):
      mSM->a->value = k_new_f ( conv_sf (input) );
      break;
    case (tStr):
      mSM->a->value = k_clone (input);
      break;
  }
}

//// mRunConsoleOutput
void mRunConsoleOutput ()
{
  switch (mSM->a->type)
  {
    case (tInt):
      printf ("%d", *(int *)mSM->a->value);
      break;
    case (tRat):
      printf ("%f", (double) *(float *)mSM->a->value);
      break;
    case (tStr):
      printf ("%s", (char *)mSM->a->value);
      break;
  }
}

//// mRunASCIIConversion
void mRunASCIIConversion ()
{
  // FORMAT: asc (dest:str) (source:int) (null) 0 //

  char source;
  char *dest;

  source = (char) *(int *)mSM->b->value;
  dest = k_new (2);
  dest[0] = source;
  dest[1] = '\0';

  JFREE(mSM->a->value); // free the current value
  mSM->a->value = dest;
}

//// mRunRandom
void mRunRandom ()
{
  int num;

  JFREE(mSM->a->value); // free the current value

  num = rand();

  switch (mSM->a->type)
  {
    case (tInt):
      mSM->a->value = k_new_n ( num );
      break;
  }
}

//// mRunFileIO
void mRunFileIO ()
{
  // fio IO_TYPE:int B C
  // IO_TYPE   OPERATION     B                   C 
  // -------   -----------   -----------------   --------------
  // 0         open input    dest handle (int)   filename (str)
  // 1         open output   dest handle (int)   filename (str)
  // 2         close         handle (int)        NULL
  // 3         read          handle (int)        dest line (str)
  // 4         write         handle (int)        line (str)

  int i;
  char tempbuf[GLOBAL_RE_FIO_LINE_LEN];

  switch (*(int *)mSM->a->value)
  {
    case 0: // open input
    {
      for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
      {
        if (mFiles[i] == NULL) break;
      }
      if (i == GLOBAL_RE_MAX_FILE_HANDLES)
      {
        err (true, ER_MAX_FILES_OPENED, NULL);
        erre (EFF_TERMINATE);
      }

      JFREE(mSM->b->value); // free the current value
      mSM->b->value = k_new_n (i);

      mFiles[i] = (FileElement *) JALLOC(sizeof (FileElement));
      mFiles[i]->file = fopen ((char *)mSM->c->value, "r");
      mFiles[i]->mode = fioInput;

      if (!(mFiles[i]->file))
      {
        err (true, ER_FILE_OPEN_ERROR, (char *)(mSM->c->value));
        erre (EFF_TERMINATE);
      }

      break;
    }
    case 1: // open output
    {
      for (i = 0; i < GLOBAL_RE_MAX_FILE_HANDLES; i++)
      {
        if (mFiles[i] == NULL) break;
      }
      if (i == GLOBAL_RE_MAX_FILE_HANDLES)
      {
        err (true, ER_MAX_FILES_OPENED, NULL);
        erre (EFF_TERMINATE);
      }

      JFREE(mSM->b->value); // free the current value
      mSM->b->value = k_new_n (i);

      mFiles[i] = (FileElement *) JALLOC(sizeof (FileElement));
      mFiles[i]->file = fopen ((char *)mSM->c->value, "w");
      mFiles[i]->mode = fioOutput;

      if (!(mFiles[i]->file))
      {
        err (true, ER_FILE_OPEN_ERROR, (char *)(mSM->c->value));
        erre (EFF_TERMINATE);
      }

      break;
    }
    case 2: // close
    {
      i = *(int *)mSM->b->value;
      if (mFiles[i] != NULL)
      {
        fclose (mFiles[i]->file);
        JFREE(mFiles[i]);
        mFiles[i] = NULL;
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
    case 3: // read
    {
      i = *(int *)mSM->b->value;
      if (mFiles[i] != NULL)
      {
        if (mFiles[i]->mode == fioInput)
        {
          JFREE(mSM->c->value);
          fgets (tempbuf, GLOBAL_RE_FIO_LINE_LEN, mFiles[i]->file);
          mSM->c->value = k_clone (tempbuf);
        }
        else
        {
          err (true, ER_INVALID_FILE_MODE, NULL);
          erre (EFF_IGNORE);
        }
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
    case 4: // write
    {
      i = *(int *)mSM->b->value;
      if (mFiles[i] != NULL)
      {
        if (mFiles[i]->mode == fioOutput)
        {
          fprintf (mFiles[i]->file, "%s", (char*)mSM->c->value);
        }
        else
        {
          err (true, ER_INVALID_FILE_MODE, NULL);
          erre (EFF_IGNORE);
        }
      }
      else
      {
        err (true, ER_INVALID_FILE_HANDLE, NULL);
        erre (EFF_IGNORE);
      }
      break;
    }
  }
}

//////////////
// Special  //
//////////////

void mGetStatement ()
{
  mSM = icm_getnext();
  if (mSM) mTAC = mSM->tac;
  else
  {
    err (true, ER_END_OF_FILE, NULL);
    errm ("Program must be terminated with a \"stop\" instruction", NULL);
    erre (EFF_TERMINATE);
  }
}

#endif

